#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import yaml
from pathlib import Path


def http_source(config):
    if 'url' not in config:
        raise TypeError('Kind http-source requires a url.')
    url = config['url']
    period = config.get('period', 1000)
    return {'uri': f'timer:tick?period={period}', 'steps': [{'to': url}]}


def _kafka_SASL_uri(topic, kafka_brokers, password):
    security_protocol = 'SASL_SSL'
    username = "token"
    conf = 'org.apache.kafka.common.security.plain.PlainLoginModule' \
        f' required username=\"{username}\" password=\"{password}\";'

    if password == 'modeline':
        return f'kafka:{topic}?brokers={kafka_brokers}' \
               f'&securityProtocol={security_protocol}' \
               f'&saslMechanism=PLAIN'
    return f'kafka:{topic}?brokers={kafka_brokers}' \
           f'&securityProtocol={security_protocol}' \
           f'&saslMechanism=PLAIN' \
           f'&saslJaasConfig={conf}' \
           f'&sslKeyPassword={password}'


def kafka_source(config):
    if 'topic' not in config:
        raise TypeError('Kind kafka-source requires a topic.')
    if 'brokers' not in config:
        raise TypeError('Kind kafka-source requires a valid kafka broker.')
    topic = config['topic']
    kafka_brokers = config['brokers']
    uri = f'kafka:{topic}?brokers={kafka_brokers}'
    if 'SASL_password' in config:
        uri = _kafka_SASL_uri(topic, kafka_brokers, config['SASL_password'])
    return {'uri': uri, 'steps': []}


def telegram_source(config):
    if 'authorization_token' not in config:
        raise TypeError('Telegram source requires an authorization token.')
    authorization_token = config['authorization_token']
    return {
        'uri': f'telegram:bots?authorizationToken={authorization_token}',
        'steps': [{
            'marshal': {
                'json': {}
            }
        }]
    }


def _binance_route(period, coin):
    return {
        'uri':
        f"timer:update?period={period}",
        'steps': [{
            'to':
            "xchange:binance?service=marketdata&method=ticker&"
            f"currencyPair={coin}/USDT"
        }, {
            'marshal': {
                'json': {}
            }
        }]
    }


def binance_source(config):
    if 'coin' not in config:
        raise TypeError(
            'Crypto source requires the official cryptocurrency symbol'
            '(example, for Bitcoin the symbol is BTC).')
    coin = config['coin']
    period = 3000
    if 'period' in config:
        period = config['period']

    if isinstance(coin, list):
        routes = []
        for coin_id in coin:
            routes.append(_binance_route(period, coin_id))
        return routes

    return _binance_route(period, coin)


def cos_source(config):
    if 'bucket_name' not in config:
        raise TypeError('Cloud object storage source requires a bucket name.')
    if 'access_key_id' not in config:
        raise TypeError(
            'Cloud object storage source requires an access key id.')
    if 'secret_access_key' not in config:
        raise TypeError(
            'Cloud object storage source requires an secret access key.')
    if 'endpoint' not in config:
        raise TypeError('Cloud object storage source requires an endpoint.')
    bucket_name = config['bucket_name']
    access_key_id = config['access_key_id']
    secret_access_key = config['secret_access_key']
    endpoint = config['endpoint']

    # Ensure this is a valid, supported endpoint:
    split_endpoint = _parse_endpoint(endpoint)

    # Resolve region:
    region = 'us-east'
    if 'region' in config:
        region = config['region']
    else:
        region = split_endpoint[1]

    # Assemble URI:
    uri = f'aws2-s3://{bucket_name}?accessKey={access_key_id}' \
          f'&secretKey={secret_access_key}' \
          '&overrideEndpoint=true' \
          f'&uriEndpointOverride={endpoint}' \
          f'&region={region}'

    # Move after read options:
    if 'move_after_read' in config:
        # Autocreation is not supported for COS:
        uri += '&autoCreateBucket=false'
        uri += '&moveAfterRead=true'

        new_bucket_name = config['move_after_read']
        uri += f'&destinationBucket={new_bucket_name}'

    # Only send an event that data is available do not send the
    # actual data.
    if 'meta_event_only' in config and config['meta_event_only']:
        uri += '&includeBody=false'
        uri += '&autocloseBody=true'

    return {'uri': uri, 'steps': []}


def file_source(config):
    if 'path' not in config:
        raise TypeError('File source requires a path name.')
    path = Path(config['path'])
    if path.is_dir():
        uri = f'file:{str(path)}?'
    else:
        uri = f'file:{str(path.parent)}?filename={path.name}&'

    # Keep files after being processed, default is true.
    if 'keep_files' not in config or not config['keep_files']:
        uri += 'delete=true'
    else:
        uri += 'delete=false'

    # Recursive traversal of the directory, default is false.
    if 'recursive' in config and config['recursive']:
        uri += '&recursive=true'

    return {'uri': uri, 'steps': []}


def file_watch_source(config):
    if 'path' not in config:
        raise TypeError('File watch source requires a directory path name.')
    path = Path(config['path'])
    if not path.is_dir():
        raise RuntimeError(f'Path {str(path)} is not a directory.')

    uri = f'file-watch:{str(path)}'

    # Set up the events to be monitored. Use the events set by the user
    # otherwise enable all valid events.
    valid_events = ["DELETE", "CREATE", "MODIFY"]
    if 'events' in config:
        events = config['events']
        event_types = events.split(",")
        for event in event_types:
            if event not in valid_events:
                raise RuntimeError(f'Event type {event} is not supported,'
                                   ' must be one of: DELETE, CREATE, MODIFY')
        uri += f'?events={events}'
    else:
        uri += f'?events={",".join(valid_events)}'

    # Recursive traversal of the directory, default is false.
    if 'recursive' in config and config['recursive']:
        uri += '&recursive=true'

    uri += "&autoCreate=false"

    return {'uri': uri, 'steps': []}


def generic_source(config):
    if 'spec' in config:
        spec = config['spec']
        if isinstance(config['spec'], str):
            spec = _process_generic_spec_str(config)
    elif 'uri' in config:
        # A source URI has been passed in, incorporate it in the generic
        # source YAML.
        spec = {'uri': config['uri'], 'steps': []}
    else:
        raise TypeError('Kind generic-source requires a spec or a uri field.')

    # If the source spec is given as non-string we assume it is a valid
    # dictionary of the form:
    # {'uri':<uri_value>, 'steps':[<more yaml or empty>]}
    if 'uri' not in spec:
        raise TypeError('Generic source requires a uri.')
    if 'steps' not in spec:
        raise TypeError('Steps field missing.')

    return spec


def generic_periodic_source(config):
    period = config.get('period', 1000)

    if 'spec' in config:
        # Parse string config if present otherwise spec is config:
        spec = config['spec']
        if isinstance(config['spec'], str):
            spec = _process_generic_spec_str(config)
    elif 'uri' in config:
        spec = {'uri': config['uri'], 'steps': []}
    else:
        raise TypeError('Kind generic-source requires a spec or a uri field.')

    # The target uri is the uri we will periodically poll.
    if 'uri' not in spec:
        raise TypeError('Periodic generic source requires a uri.')
    target_uri = spec['uri']

    periodic_spec = {
        'uri': f'timer:tick?period={period}',
        'steps': [{
            'to': target_uri
        }]
    }

    # Append any additional steps:
    periodic_spec['remaining_steps'] = spec['steps']

    return periodic_spec


sources = {
    'http-source': http_source,
    'kafka-source': kafka_source,
    'telegram-source': telegram_source,
    'binance-source': binance_source,
    'cloud-object-storage-source': cos_source,
    'file-source': file_source,
    'file-watch-source': file_watch_source,
    'generic-source': generic_source,
    'generic-periodic-source': generic_periodic_source
}


def _finalize_route(spec, endpoint, inverted, add_to_queue):
    if inverted:
        spec['steps'].append({'bean': add_to_queue})
    else:
        spec['steps'].append({'to': endpoint})
    spec = [{'from': spec}]
    return spec


# Construct a camel source specification from a rayvens source config
def construct_source(config, endpoint, inverted=False):
    if 'kind' not in config:
        raise TypeError('A Camel source needs a kind.')
    kind = config['kind']
    source_handler = sources.get(kind)
    if source_handler is None:
        raise TypeError(f'Unsupported Camel source: {kind}.')

    spec = source_handler(config)

    # Extract remaining steps if any:
    remaining_steps = None
    if 'remaining_steps' in spec:
        remaining_steps = spec['remaining_steps']
        del spec['remaining_steps']

    # Manage Queue access methods.
    take_from_queue = 'takeFromQueue'
    add_to_queue = 'addToQueue'
    if config['kind'] == 'file-source':
        take_from_queue = 'takeFromFileQueue'
        add_to_queue = 'addToFileQueue'
    elif config['kind'] == 'file-watch-source':
        take_from_queue = 'takeFromFileWatchQueue'
        add_to_queue = 'addToFileWatchQueue'
    elif config['kind'] == 'cloud-object-storage-source' and \
            'meta_event_only' in config and config['meta_event_only']:
        take_from_queue = 'takeFromMetaEventQueue'
        add_to_queue = 'addToMetaEventQueue'
    elif config['kind'] == 'cloud-object-storage-source':
        take_from_queue = 'takeFromFileJsonQueue'
        add_to_queue = 'addToFileJsonQueue'

    # Multi-source integration with several routes:
    if isinstance(spec, list):
        # Generic sources do now allow multiple routes:
        if remaining_steps is not None:
            raise TypeError(
                'Generic source with multiple routes not supported')

        spec_list = []
        for spec_entry in spec:
            spec_list.extend(
                _finalize_route(spec_entry, endpoint, inverted, add_to_queue))
        if inverted:

            spec_list.append({
                'from': {
                    'uri': endpoint,
                    'steps': [{
                        'bean': take_from_queue
                    }]
                }
            })
        # print(yaml.dump(spec_list))
        return spec_list

    # Regular integration with only one route:
    spec = _finalize_route(spec, endpoint, inverted, add_to_queue)
    if inverted:
        # Route fetching from queue:
        from_queue = {
            'from': {
                'uri': endpoint,
                'steps': [{
                    'bean': take_from_queue
                }]
            }
        }

        # Attach remaining steps after fetching from the queue.
        if remaining_steps is not None:
            from_queue['from']['steps'].extend(remaining_steps)

        spec.append(from_queue)
    # print(yaml.dump(spec))
    return spec


def slack_sink(config):
    if 'channel' not in config:
        raise TypeError('Kind slack-sink requires a channel.')
    channel = config['channel']

    option_values = []
    if 'webhook_url' in config:
        webhook_url = config['webhook_url']
        option_values.append(f"webhookUrl={webhook_url}")

    options = ""
    if len(option_values) > 0:
        options = "?" + "&".join(option_values)

    final_spec = {
        'steps': [{
            'to': f'slack:{channel}{options}',
        }]
    }

    return [(final_spec, None)]


def kafka_sink(config):
    if 'topic' not in config:
        raise TypeError('Kind kafka-sink requires a topic.')
    if 'brokers' not in config:
        raise TypeError('Kind kafka-sink requires a valid kafka broker.')
    topic = config['topic']
    kafka_brokers = config['brokers']
    to = f'kafka:{topic}?brokers={kafka_brokers}'
    if 'SASL_password' in config:
        to = _kafka_SASL_uri(topic, kafka_brokers, config['SASL_password'])

    final_spec = {
        'steps': [{
            'to': to,
        }]
    }
    return [(final_spec, None)]


def telegram_sink(config):
    if 'authorization_token' not in config:
        raise TypeError('Authorization token required for telegram sink.')
    authorization_token = config['authorization_token']
    chat_id = ""
    if 'chat_id' in config:
        chat_id = config['chat_id']

    # The telegram sink yaml below allows to override the default chat ID where
    # the messages are sent.
    final_spec = {
        'steps': [{
            'convert-body-to': {
                'type': "java.lang.String"
            }
        }, {
            'to': {
                'uri': 'telegram:bots',
                'parameters': {
                    'authorizationToken': authorization_token,
                    'chatId': chat_id
                }
            }
        }, {
            'marshal': {
                'json': {}
            }
        }]
    }
    return [(final_spec, None)]


def cos_sink(config):
    if 'bucket_name' not in config:
        raise TypeError('Cloud object storage sink requires a bucket name.')
    if 'access_key_id' not in config:
        raise TypeError('Cloud object storage sink requires an access key id.')
    if 'secret_access_key' not in config:
        raise TypeError(
            'Cloud object storage sink requires an secret access key.')
    if 'endpoint' not in config:
        raise TypeError('Cloud object storage sink requires an endpoint.')
    bucket_name = config['bucket_name']
    access_key_id = config['access_key_id']
    secret_access_key = config['secret_access_key']
    endpoint = config['endpoint']

    file_name = "default.txt"
    if 'file_name' in config:
        file_name = config['file_name']

    # Ensure this is a valid, supported endpoint:
    split_endpoint = _parse_endpoint(endpoint)

    # Resolve region:
    region = 'us-east'
    if 'region' in config:
        region = config['region']
    else:
        region = split_endpoint[1]

    # If we are uploading a file either directly or by monitoring a directory,
    # multi-part must be enabled:
    uploads_file = 'from_file' in config or 'from_directory' in config
    if uploads_file:
        config['upload_type'] = "multi-part"

    # Assemble URI:
    uri = f'aws2-s3://{bucket_name}?accessKey={access_key_id}' \
          f'&secretKey={secret_access_key}' \
          '&overrideEndpoint=true' \
          f'&uriEndpointOverride={endpoint}' \
          f'&region={region}'

    # Final result is a list of spec and uri pairs: [(spec, uri)]
    spec_list = []

    # Streaming is only supported in latest Camel. Camel-K currently
    # supports Camel 3.9.0 only. Camel 3.10 or newer is required for
    # this feature.
    if 'upload_type' in config:
        if config['upload_type'] == "stream":
            uri += '&streamingUploadMode=true'
            uri += '&namingStrategy=progressive'
            uri += '&restartingPolicy=lastPart'
            # 1 MB sized batches are default in streaming mode.
            batch_size = 1000000  # bytes
            if 'batch_size' in config:
                batch_size = config['batch_size']
            uri += f'&batchSize={batch_size}'

            # 10 messages per batch is default in upload mode.
            messages_per_batch = 10
            if 'messages_per_batch' in config:
                messages_per_batch = config['messages_per_batch']
            uri += f'&batchMessageNumber={messages_per_batch}'
            raise TypeError("Streaming uploads not yet supported.")
        elif config['upload_type'] == "multi-part":
            uri += '&multiPartUpload=true'
            # 25 MB size per part is the default in streaming mode.
            part_size = 26214400  # bytes
            if 'part_size' in config:
                part_size = config['part_size']
            uri += f'&partSize={part_size}'

            # Initialize the part of the spec that has its source in Rayvens,
            # i.e. the part which takes as input file names from the user.
            spec = {'steps': []}
            spec['steps'].append({'bean': 'processPath'})
            # Overwrite existing file name if user provided a new name.
            if file_name is not None:
                spec['steps'].append({
                    'set-header': {
                        'name': 'CamelAwsS3Key',
                        'simple': file_name
                    }
                })
            spec['steps'].append({'to': uri})
            spec_list.append((spec, None))

            # If the from_file option is active we need to create a route
            # from the local file to the cloud object storage. For this route
            # we do not allow the overwriting of the original file name since
            # that may cause a clash with the file uploaded on the Rayvens-to-
            # COS route above for which name overwriting is supported.
            # When from_directory option is active all the files dumped into
            # a file system directory will be uploaded to Cloud Object Storage.
            # `from_file` and `from_directory` cannot be active at the same
            # time.
            if 'from_file' in config:
                # Process file path and create from_uri:
                from_file_path = Path(config['from_file'])
                uploaded_file_name = from_file_path.name
                file_dir = str(from_file_path.parent)
                from_uri = f'file:{file_dir}?filename={uploaded_file_name}'
                # Delete file after it is uploaded to avoid the files being
                # copied to a temporary folder after being uploaded.
                if 'keep_file' in config and config['keep_file']:
                    from_uri += '&delete=false'
                else:
                    from_uri += '&delete=true'

                # Create new file route for final spec:
                file_spec = {'steps': []}
                file_spec['steps'].append({'bean': 'processFile'})
                file_spec['steps'].append({'to': uri})
                spec_list.append((file_spec, from_uri))
            elif 'from_directory' in config:
                # Process file path and create from_uri:
                from_directory_path = config['from_directory']
                from_uri = f'file:{from_directory_path}'
                # Delete file after it is uploaded to avoid the files being
                # copied to a temporary folder after being uploaded.
                if 'keep_file' in config and config['keep_file']:
                    from_uri += '?delete=false'
                else:
                    from_uri += '?delete=true'

                # Create new file route for final spec:
                file_spec = {'steps': []}
                file_spec['steps'].append({'bean': 'processFile'})
                file_spec['steps'].append({'to': uri})
                spec_list.append((file_spec, from_uri))
            return spec_list
        else:
            raise TypeError(
                "Unrecognized upload type. Use one of: stream, multi-part.")

    # This is the default behavior when user application data is
    # written into a COS file directly.
    if 'file_name' in config:
        regular_spec = {
            'steps': [{
                'set-header': {
                    'name': 'CamelAwsS3Key',
                    'simple': f"{file_name}"
                }
            }, {
                'to': uri
            }]
        }
    else:
        regular_spec = {'steps': [{'to': uri}]}
    spec_list.append((regular_spec, None))
    return spec_list


def test_sink(config):
    return [({'steps': [{'log': {'message': "\"${body}\""}}]}, None)]


def generic_sink(config):
    if 'spec' in config:
        if isinstance(config['spec'], str):
            return [(_process_generic_spec_str(config, sink=True), None)]
        # If the sink spec is given as non-string we assume it is a valid
        # dictionary of the form:
        # {'uri':<uri_value>, 'steps':[<more yaml or empty>]}
        return [(config['spec'], None)]
    elif 'uri' in config:
        return [({'steps': [{'to': config['uri']}]}, None)]
    else:
        raise TypeError('Kind generic-sink requires a spec or uri field.')


sinks = {
    'slack-sink': slack_sink,
    'kafka-sink': kafka_sink,
    'telegram-sink': telegram_sink,
    'cloud-object-storage-sink': cos_sink,
    'generic-sink': generic_sink,
    'test-sink': test_sink
}


# Construct a camel sink specification from a rayvens sink config.
def construct_sink(config, endpoint):
    if 'kind' not in config:
        raise TypeError('A Camel sink needs a kind.')
    kind = config['kind']
    sink_handler = sinks.get(kind)
    if sink_handler is None:
        raise TypeError(f'Unsupported Camel sink: {kind}.')

    spec_list = sink_handler(config)

    # Add the from uri construct to each route and finalize spec.
    final_spec_list = []
    for spec_uri_pair in spec_list:
        spec = spec_uri_pair[0]
        from_uri = spec_uri_pair[1]
        if from_uri is None:
            spec['uri'] = endpoint
        else:
            spec['uri'] = from_uri
        final_spec_list.append({'from': spec})
    # print(yaml.dump(final_spec_list))
    return final_spec_list


def no_restriction(config):
    return dict(restricted_message_types=[])


def cos_sink_restriction(config):
    # The input type for this sink is a file denoted by the type Path.
    if 'upload_type' in config and config['upload_type'] == 'multi-part' and \
       'from_directory' not in config:
        return dict(restricted_message_types=[Path])
    return no_restriction(config)


sink_input_restriction = {
    'slack-sink': no_restriction,
    'kafka-sink': no_restriction,
    'telegram-sink': no_restriction,
    'cloud-object-storage-sink': cos_sink_restriction,
    'generic-sink': no_restriction,
    'test-sink': no_restriction
}


# Return a dict of valid message types
def input_restriction(config):
    if 'kind' not in config:
        raise TypeError('A Camel sink needs a kind.')
    handler = sink_input_restriction.get(config['kind'])
    return handler(config)


def _process_generic_spec_str(config, sink=False):
    # Parse string to a Python dictionary.
    generic_spec = yaml.safe_load(config['spec'])

    # Ensure only a single source exists.
    # TODO: enable support for multi-source Yaml.
    if isinstance(generic_spec, list):
        if len(generic_spec) > 1:
            raise TypeError('Generic spec field has multiple sources.')
        generic_spec = generic_spec[0]

    # from will be inserted later.
    if 'from' in generic_spec:
        if generic_spec['from'] is None:
            del generic_spec['from']
        else:
            generic_spec = generic_spec['from']

    # A uri for the source must be present.
    if not sink and 'uri' not in generic_spec:
        raise TypeError('Generic spec needs a uri entry.')

    # A steps field is required but can be empty by default.
    if 'steps' not in generic_spec:
        generic_spec['steps'] = []

    return generic_spec


def _parse_endpoint(endpoint):
    # Check URI scheme is valid:
    if endpoint.startswith("https://"):
        trimmed_endpoint = endpoint[len("https://"):]
        split_endpoint = trimmed_endpoint.split(".")

        if split_endpoint[0] != "s3":
            raise TypeError(f"Endpoint {endpoint} is not an s3 endpoint.")

        return split_endpoint

    raise TypeError(
        f"Unexpected or missing URI scheme in endpoint: {endpoint}")



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import ray
import time

from rayvens.core.local import start as start_http
from rayvens.core.kafka import start as start_kafka
from rayvens.core.operator import start as start_operator_http
from rayvens.core.operator_kafka import start as start_operator_kafka
from rayvens.core.ray_serve import start as start_operator_ray_serve
from rayvens.core.name import name_source, name_sink
from rayvens.core.verify import verify_do
from queue import Queue


class Stream:
    def __init__(self,
                 name,
                 actor_options=None,
                 operator=None,
                 source_config=None,
                 sink_config=None,
                 batch_size=None):
        if _global_camel is None:
            raise RuntimeError(
                "Rayvens has not been started. Start with 'rayvens.init()'.")
        self.name = name
        self.actor = StreamActor.options(actor_options).remote(
            name, operator=operator, batch_size=batch_size)
        if sink_config is not None:
            self.add_sink(sink_config)
        if source_config is not None:
            self.add_source(source_config)

    def send_to(self, subscriber, name=None):
        ray.get(self.actor.send_to.remote(subscriber, name))
        return subscriber

    def append(self, data):
        self.actor.append.remote(data)
        return self

    def add_operator(self, operator, batch_size=None):
        return ray.get(self.actor.add_operator.remote(operator, batch_size))

    def add_multitask_operator(self, operator, batch_size=None):
        return ray.get(
            self.actor.add_multitask_operator.remote(operator, batch_size))

    def add_source(self, source_config):
        return ray.get(self.actor.add_source.remote(self, source_config))

    def add_sink(self, sink_config):
        return ray.get(self.actor.add_sink.remote(self, sink_config))

    def unsubscribe(self, subscriber_name):
        return ray.get(self.actor.unsubscribe.remote(subscriber_name))

    def disconnect_source(self, source_name, after_idle_for=None, after=None):
        self._wait_for_timeout(after_idle_for)
        return ray.get(self.actor.disconnect_source.remote(source_name))

    def disconnect_sink(self, sink_name, after_idle_for=None, after=None):
        self._wait_for_timeout(after_idle_for)
        return ray.get(self.actor.disconnect_sink.remote(sink_name))

    def disconnect_all(self,
                       after_idle_for=None,
                       after=None,
                       stream_drain_timeout=5):
        self._wait_for_timeout(after_idle_for, after)
        return ray.get(self.actor.disconnect_all.remote(stream_drain_timeout))

    def event_count(self):
        return ray.get(self.actor.event_count.remote())

    def _meta(self, action, *args, **kwargs):
        return ray.get(self.actor._meta.remote(action, *args, **kwargs))

    def _wait_for_timeout(self, after_idle_for, after):
        if after_idle_for is not None and after_idle_for > 0:
            while True:
                time_elapsed_since_last_event = self._idle_time()

                if time_elapsed_since_last_event is not None:
                    # Idle timeout exceeds the user-specified time limit:
                    if time_elapsed_since_last_event > after_idle_for:
                        break

                    # Check again after waiting for the rest of the timeout
                    # time:
                    time.sleep(after_idle_for - time_elapsed_since_last_event +
                               1)
                else:
                    time.sleep(after_idle_for)
        if after is not None and after > 0:
            time.sleep(after)

    def _idle_time(self):
        latest_timestamp = ray.get(self.actor._get_latest_timestamp.remote())
        if latest_timestamp is None:
            return None
        return time.time() - latest_timestamp


class StreamContext:
    def __init__(self, stream_batch_size):
        self.sink_restrictions = {}
        self.subscribers = {}
        self.latest_sent_event_timestamp = None
        self.event_counter = 0
        self.limit_subscribers = False
        self.is_multi_operator = False
        self.stream_batch_size = stream_batch_size

    def publish(self, data):
        if data is not None:
            for name, subscriber in self.subscribers.items():
                if name in self.sink_restrictions:
                    type_restrictions = self.sink_restrictions[name]
                    if not self._accepts_data_type(data, type_restrictions):
                        continue
                _eval(self, subscriber, data)

    # Check if the sink we are routing the message to has any restrictions
    # in terms of message type. A message will only be routed to a sink
    # if the sink accepts its type.
    def _accepts_data_type(self, data, type_restrictions):
        # If there are no restrictions return immediately:
        if len(type_restrictions) == 0:
            return True
        for restricted_type in type_restrictions:
            if isinstance(data, restricted_type):
                return True
        return False


@ray.remote(num_cpus=0)
class StreamActor:
    def __init__(self, name, operator=None, batch_size=None):
        self.name = name
        self._operator = operator
        self._sources = {}
        self._sinks = {}
        self.context = StreamContext(batch_size)
        self.queue = None
        if self.context.stream_batch_size is not None:
            self.queue = Queue(maxsize=self.context.stream_batch_size)

    def send_to(self, subscriber, name=None):
        if self.context.limit_subscribers:
            return
        if name in self.context.subscribers:
            raise RuntimeError(
                f'Stream {self.name} already has a subscriber named {name}.')
        if name is None:
            name = object()
        self.context.subscribers[name] = subscriber

    def append(self, data):
        if data is None:
            return
        if self.queue is not None:
            self.queue.put_nowait(data)
            if not self.queue.full():
                return
            data = list(self.queue.queue)
            with self.queue.mutex:
                self.queue.queue.clear()
        if self._operator is not None:
            data = _eval(self.context, self._operator, data)
        self.context.publish(data)
        self.context.latest_sent_event_timestamp = time.time()
        self.context.event_counter += 1

    def add_operator(self, operator, batch_size):
        self._operator = operator
        self.context.batch_size = batch_size

    def add_multitask_operator(self, operator, batch_size):
        self._operator = operator
        self.context.batch_size = batch_size
        self.context.is_multi_operator = True

    def add_source(self, stream, source_config):
        source_config["integration_type"] = 'source'
        source_name = name_source(source_config)
        if source_name in self._sources:
            raise RuntimeError(
                f'Stream {self.name} already has a source named {source_name}.'
            )
        self._sources[source_name] = _global_camel.add_source(
            stream, source_config, source_name)
        return source_name

    def add_sink(self, stream, sink_config):
        sink_config["integration_type"] = 'sink'
        sink_name = name_sink(sink_config)
        if sink_name in self._sinks:
            raise RuntimeError(
                f'Stream {self.name} already has a sink named {sink_name}.')
        self._sinks[sink_name] = _global_camel.add_sink(
            stream, sink_config, sink_name)
        self.context.sink_restrictions[sink_name] = self._sinks[
            sink_name].get_restricted_data_type()
        return sink_name

    def unsubscribe(self, subscriber_name):
        if subscriber_name not in self._subscribers:
            raise RuntimeError(f'Stream {self.name} has no subscriber named'
                               f' {subscriber_name}.')
        self.context.subscribers.pop(subscriber_name)

    def disconnect_source(self, source_name):
        if source_name not in self._sources:
            raise RuntimeError(
                f'Stream {self.name} has no source named {source_name}.')
        _global_camel.disconnect(self._sources[source_name])
        self._sources.pop(source_name)

    def disconnect_sink(self, sink_name):
        if sink_name not in self._sinks:
            raise RuntimeError(
                f'Stream {self.name} has no sink named {sink_name}.')
        _global_camel.disconnect(self._sinks[sink_name])
        self._sinks.pop(sink_name)
        self.context.sink_restrictions.pop(sink_name)
        self.context.subscribers.pop(sink_name)

    def disconnect_all(self, stream_drain_timeout):
        for source_name in dict(self._sources):
            self.disconnect_source(source_name)
        if self.queue is not None:
            self.flush_batch()
        time.sleep(stream_drain_timeout)
        for sink_name in dict(self._sinks):
            self.disconnect_sink(sink_name)

    def event_count(self):
        return self.context.event_counter

    def flush_batch(self):
        data = None
        if self.queue is not None:
            data = list(self.queue.queue)
            with self.queue.mutex:
                self.queue.queue.clear()
        if data is None:
            return
        if self._operator is not None:
            data = _eval(self.context, self._operator, data)
        self.context.publish(data)
        self.context.latest_sent_event_timestamp = time.time()
        self.context.event_counter += 1

    def _get_integration(self, source_sink_name):
        if source_sink_name in self._sinks:
            integration = self._sinks[source_sink_name]
        if source_sink_name in self._sources:
            integration = self._sources[source_sink_name]
        if integration is None:
            raise RuntimeError(
                f'{source_sink_name} not found on stream {self.name}')
        return integration

    def _integration_invoke(self, source_sink_name, message):
        integration = self._get_integration(source_sink_name)
        return integration.invocation.invoke(message)

    def _get_integration_name(self, source_sink_name):
        return self._get_integration(source_sink_name).integration_name

    def _get_latest_timestamp(self):
        return self.context.latest_sent_event_timestamp

    def _fetch_processors(self):
        self.context.limit_subscribers = True
        return self.context.subscribers, self._operator

    def _update_timestamp(self, timestamp):
        self.context.latest_sent_event_timestamp = timestamp


def _eval(context, f, data):
    if isinstance(f, Stream):
        return f.append(data)
    elif isinstance(f, ray.actor.ActorHandle):
        return f.append.remote(data)
    elif isinstance(f, ray.actor.ActorMethod):
        return f.remote(data)
    elif isinstance(f, ray.remote_function.RemoteFunction):
        if context.is_multi_operator:
            if context.subscribers is None:
                raise RuntimeError('No subscribers or sinks provided.')
            f.remote(context, data)
            return None
        return f.remote(data)
    else:
        return f(data)


setattr(Stream, '__rshift__', Stream.send_to)
setattr(Stream, '__lshift__', Stream.append)

_global_camel = None


def init(mode=os.getenv('RAYVENS_MODE', 'auto'),
         transport=os.getenv('RAYVENS_TRANSPORT', 'auto'),
         release=os.getenv('RAYVENS_RELEASE', False),
         check_port=True):
    modes = ['auto', 'local', 'mixed', 'operator']
    transports = ['auto', 'http', 'kafka', 'ray-serve']

    if mode not in modes:
        raise RuntimeError(
            f'Unsupported Rayvens mode. Must be one of {modes}.')
    if transport not in transports:
        raise RuntimeError(
            f'Unsupported Rayvens transport. Must be one of {transports}.')

    global _global_camel

    if mode in ['auto', 'local']:
        if transport in ['auto', 'http']:
            _global_camel = start_http(mode, check_port, release)
        elif transport == 'kafka':
            _global_camel = start_kafka(mode, check_port, release)
        else:
            raise RuntimeError(
                f'{transport} transport unsupported for mode {mode}.')
    elif mode in ['mixed', 'operator']:
        if transport in ['auto', 'http']:
            _global_camel = start_operator_http(mode, check_port, release)
        elif transport == 'kafka':
            _global_camel = start_operator_kafka(mode, check_port, release)
        elif transport in ['ray-serve']:
            _global_camel = start_operator_ray_serve(mode, check_port, release)
        else:
            raise RuntimeError(
                f'{transport} transport unsupported for mode {mode}.')
    else:
        raise RuntimeError(f'Unsupported mode {mode}.')


def meta(stream, action, *args, **kwargs):
    return verify_do(stream.actor, _global_camel, action, *args, **kwargs)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import yaml
import rayvens.cli.utils as utils
import rayvens.cli.file as file
import rayvens.cli.java as java
import rayvens.cli.docker as docker
from rayvens.core.catalog import sources, sinks
from rayvens.core.catalog import construct_source, construct_sink


def build_base_image(args):
    # Create docker file for the base image.
    docker_image = docker.JavaAlpineDockerImage()

    # Install packages into the image.
    docker_image.install("maven")
    docker_image.install("bash")
    docker_image.install("curl")
    docker_image.update_installed_packages()

    # Bring in kamel executable:
    docker_image.add_kamel()

    # TODO: remove this, overwrite kamel executable with one from host.
    path_to_local_kamel = file.find_executable("kamel-linux")
    kamel_executable = file.File(path_to_local_kamel)
    docker_image.copy(kamel_executable, "/usr/local/bin/kamel")

    # Add kubernetes capabilities:
    docker_image.add_kubernetes()

    # Create preloader file with string content type:
    preload_file = file.File(java.preloader_file_name,
                             contents=java.preloader_file_contents)
    docker_image.copy(preload_file)

    # Add run command:
    docker_image.run(f"""kamel local run {preload_file.name} \
--dependency mvn:org.apache.camel.quarkus:camel-quarkus-java-joor-dsl \
--dependency camel:camel-quarkus-microprofile-health; \
rm {preload_file.name}""")

    # Build image:
    docker_image.build(utils.get_base_image_name(args))

    # Push base image to registry.
    docker_image.push()


def build_integration(args):
    # Create image for integration:
    docker_image = docker.DockerImage(utils.get_base_image_name(args))

    # Create workspace inside the image.
    docker_image.run(f"mkdir -p /{docker.image_workspace_name}")
    docker_image.workdir(f"/{docker.image_workspace_name}")

    # Create the yaml source for the integration locally and transfer it
    # to the image.
    predefined_integration = args.kind is not None and (args.kind in sources
                                                        or args.kind in sinks)
    inverted_transport = True
    launches_kubectl_jobs = args.launch_image is not None

    # Put together the summary file.
    summary_file = utils.get_summary_file(args)

    # The name of the integration:
    name = args.kind
    if args.name is not None:
        name = args.name

    # The name of the integration file:
    integration_file_name = utils.get_kubernetes_integration_file_name(name)

    if predefined_integration:
        # Get a skeleton configuration for this integration kind.
        base_config, _ = utils.get_current_config(args)

        # Create the integration yaml specification.
        route = "/" + args.kind + "-route"
        if args.kind in sources:
            spec = construct_source(base_config,
                                    f'platform-http:{route}',
                                    inverted=inverted_transport)
        else:
            spec = construct_sink(base_config, f'platform-http:{route}')

        # Any command line specified properties get transformed into modeline
        # options. Modeline options live at the top of the integration source
        # file.
        modeline_options = utils.get_modeline_config(args)
        integration_source_file = modeline_options + "\n\n" + yaml.dump(spec)
        integration_file = file.File(integration_file_name,
                                     contents=integration_source_file)
        docker_image.copy(integration_file)

        # Check if additional files need to be added.
        additional_files = utils.get_additional_files(spec, inverted_transport,
                                                      args.launch_image)
        for additional_file in additional_files:
            docker_image.copy(additional_file)

        # Add summary file to image.
        docker_image.copy(summary_file)
    else:
        raise RuntimeError("Not implemented yet")

    # Copy the current kubeconfig to the workspace directory:
    if launches_kubectl_jobs:
        path_to_kubeconfig = os.path.expanduser('~') + "/.kube/config"
        kubeconfig_file = file.File(path_to_kubeconfig)
        docker_image.copy(kubeconfig_file)

    # Additional files to be added to the RUN command line:
    files_list = " ".join(
        [additional_file.name for additional_file in additional_files])

    # Command that is run when building the image. This command is meant
    # to preload all dependencies to run the integration.
    run_command = f"""kamel local build {integration_file.name} {files_list} \
--integration-directory my-integration \
--dependency mvn:org.apache.camel.quarkus:camel-quarkus-java-joor-dsl \
--dependency mvn:com.googlecode.json-simple:json-simple:1.1.1 \
--dependency camel:camel-quarkus-microprofile-health"""
    if launches_kubectl_jobs:
        run_command += " --dependency mvn:io.kubernetes:client-java:11.0.0"
    docker_image.run(run_command)

    # List of all environment variables either given on the command line
    # or as part of the summary file or are part of the inner image scope
    # and are relevant to kamel local run.
    envvars = utils.get_modeline_envvars(summary_file, args)
    if launches_kubectl_jobs:
        envvars.extend([
            "PATH", "KUBERNETES_SERVICE_PORT", "KUBERNETES_PORT", "HOSTNAME",
            "JAVA_VERSION", "KUBERNETES_PORT_443_TCP_ADDR",
            "KUBERNETES_PORT_443_TCP_PORT", "KUBERNETES_PORT_443_TCP_PROTO",
            "LANG", "HTTP_SOURCE_ENTRYPOINT_PORT", "KUBERNETES_PORT_443_TCP",
            "KUBERNETES_SERVICE_PORT_HTTPS", "LC_ALL", "JAVA_HOME",
            "KUBERNETES_SERVICE_HOST", "PWD"
        ])

    # The list of envvars is of the format:
    #   --env ENV_VAR=$ENV_VAR
    outer_scope_envvars = []
    for envvar in envvars:
        outer_scope_envvars.append(f"--env {envvar}=${envvar}")
    outer_scope_envvars = " ".join(outer_scope_envvars)

    # Kubernetes does not allow us to mount a file over an existing file.
    # This means that when we deploy the integration on Kubernetes, we need
    # to mount the updated integration file to a separate location in the image
    # and then copy the image in the correct location i.e. the routes subfolder
    # to overwrite the old integration yaml file.
    docker_image.cmd(
        f"{docker.update_integration_file_in_image(integration_file_name)}; "
        f"kamel local run --integration-directory "
        f"{docker.built_integration_directory} "
        f"{outer_scope_envvars}")

    # Build image.
    docker_image.build(utils.get_integration_image(args))

    # Push integration image to registry.
    docker_image.push()



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse

from rayvens.cli.build import build_integration, build_base_image
from rayvens.cli.run import run_integration
from rayvens.cli.rayvens_setup import rayvens_setup
from rayvens.cli.rayvens_print import rayvens_print
from rayvens.cli.delete import delete

parser = argparse.ArgumentParser(description='Rayvens command line tool.',
                                 prog="rayvens")

# Add sub-parsers
subparsers = parser.add_subparsers()

# =============================
# Build base image sub-command:
# =============================
parser_base = subparsers.add_parser(
    'base', help='Build base image used for all integrations.')
parser_base.add_argument('--dev',
                         action='store_true',
                         help='Use local registry localhost:5000.')
parser_base.add_argument('-r', '--registry', help='Image registry.')
parser_base.set_defaults(func=build_base_image)

# =============================
# Build image sub-command:
# =============================
parser_build = subparsers.add_parser('build',
                                     help='Build an integration image.')
parser_build.add_argument(
    '-k', '--kind', help='Integration predefined type (e.g. slack-sink).')
parser_build.add_argument('--generic-source',
                          help='Source integration generic endpoint.')
parser_build.add_argument('--generic-sink',
                          help='Sink integration generic endpoint.')
parser_build.add_argument(
    '-p',
    '--properties',
    nargs='*',
    help='Specify the name of a property or list of properties the '
    'integration should have. The properties are specified as '
    '`-p prop1=value1 prop2=value2`')

parser_build.add_argument(
    '-e',
    '--envvars',
    nargs='*',
    help='Specify the environment variable behind an integration property '
    'specified as '
    '`-e prop1=ENV_VARIABLE_1 prop2=ENV_VARIABLE_2`')

parser_build.add_argument('--dev',
                          action='store_true',
                          help='Use local registry localhost:5000.')
parser_build.add_argument('-r', '--registry', help='Image registry.')
parser_build.add_argument('-i', '--image', help='Image name.')
parser_build.add_argument(
    '-l',
    '--launch-image',
    help='Enable the launching of a job with the provided image each time '
    'an event is received. This option only makes sense for sources.')
parser_build.add_argument('--name',
                          help='Kubernetes name used for the integration.')
parser_build.set_defaults(func=build_integration)

# =============================
# Run image sub-command:
# =============================
parser_run = subparsers.add_parser(
    'run', help='Run a previously built integration image.')
parser_run.add_argument(
    '-p',
    '--properties',
    nargs='*',
    help='Specify the name of a property or list of properties the '
    'integration should have. The properties are specified as '
    '`-p prop1=value1 prop2=value2`')
parser_run.add_argument(
    '-e',
    '--envvars',
    nargs='*',
    help='Specify the environment variable behind an integration property '
    'specified as '
    '`-e prop1=ENV_VARIABLE_1 prop2=ENV_VARIABLE_2`')

parser_run.add_argument('--dev',
                        action='store_true',
                        help='Use local registry localhost:5000.')
parser_run.add_argument('-r', '--registry', help='Image registry.')
parser_run.add_argument('-i', '--image', help='Image registry.')
parser_run.add_argument('-d',
                        '--deploy',
                        action='store_true',
                        help='Deploy integration on Kubernetes.')
parser_run.add_argument(
    '-n',
    '--namespace',
    help='Kubernetes namespace used for deploying the integration.')
parser_run.add_argument('--name', help='Integration name.')
parser_run.add_argument('--handler-image',
                        help='Name of the job image used for handling events.')
parser_run.add_argument('--handler-name',
                        help='Name of the job used for handling events.')
parser_run.add_argument('-v',
                        '--verbose',
                        action='store_true',
                        help='Verbose mode.')
parser_run.set_defaults(func=run_integration)

# =============================
# Setup sub-command:
# =============================
parser_setup = subparsers.add_parser('setup', help='Rayvens setup.')
parser_setup.add_argument('--dev', help='Setup Rayvens in dev mode.')
parser_setup.set_defaults(func=rayvens_setup)

# =============================
# Print sub-command:
# =============================
parser_print = subparsers.add_parser('print',
                                     help='List details about integrations.')
parser_print.add_argument('--all',
                          action='store_true',
                          help='List all predefind integration types.')
parser_print.add_argument('-k',
                          '--kind',
                          help='List requirements for this integration.')
parser_print.set_defaults(func=rayvens_print)

# =============================
# Delete sub-command:
# =============================
parser_delete = subparsers.add_parser('delete', help='Delete integration.')
parser_delete.add_argument('--name', help='Integration name.')
parser_delete.add_argument('--deployed',
                           action='store_true',
                           help='Delete a Kubernetes-deployed integration.')
parser_delete.add_argument('-n', '--namespace', help='Deployment namespace.')
parser_delete.add_argument('--all-jobs',
                           help="Delete all jobs with the provided base name.")
parser_delete.add_argument('-v',
                           '--verbose',
                           action='store_true',
                           help='Verbose mode.')
parser_delete.set_defaults(func=delete)

args = parser.parse_args()
args.func(args)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import rayvens.core.catalog as catalog
from rayvens.core.integration import Integration
from rayvens.core.common import get_run_mode, send_to, recv_from, await_start


def start(camel_mode, check_port, release):
    return Camel(get_run_mode(camel_mode, check_port, release))


class Camel:
    def __init__(self, mode):
        self.mode = mode
        self.mode.transport = 'http'

    def add_source(self, stream, config, source_name):
        integration = Integration(stream.name, source_name, config)
        route = integration.route(default='/source')
        spec = catalog.construct_source(config,
                                        f'platform-http:{route}',
                                        inverted=True)
        integration.prepare_environment(self.mode)
        integration.invoke_local_run(self.mode, spec)
        integration.thread = send_to(stream.actor,
                                     self.mode.server_address(integration),
                                     route)
        if not await_start(self.mode, integration):
            raise RuntimeError('Could not start source')
        return integration

    def add_sink(self, stream, config, sink_name):
        integration = Integration(stream.name, sink_name, config)
        route = integration.route(default='/sink')
        spec = catalog.construct_sink(config, f'platform-http:{route}')
        integration.prepare_environment(self.mode)
        integration.invoke_local_run(self.mode, spec)
        recv_from(stream.actor, sink_name,
                  self.mode.server_address(integration), route)
        if not await_start(self.mode, integration):
            raise RuntimeError('Could not start source')
        return integration

    def disconnect(self, integration):
        integration.disconnect(self.mode)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
from setuptools import setup

long_description = '''Rayvens augments [Ray](https://ray.io) with events. With
Rayvens, Ray applications can subscribe to event streams, process and produce
events. Rayvens leverages [Apache Camel](https://camel.apache.org) to make it
possible for data scientists to access hundreds of data services with little
effort.

For the full documentation see
[https://github.com/project-codeflare/rayvens](https://github.com/project-codeflare/rayvens).
'''

with open('scripts/rayvens-setup.sh') as f:
    version = re.findall('rayvens_version=([0-9.]+)', f.read())[0]

setup(
    name='rayvens',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['rayvens', 'rayvens.core', 'rayvens.cli'],
    package_data={'rayvens.core': ['*.java']},
    install_requires=[
        'confluent-kafka>=1.6.1', 'ray[default,serve,k8s]>=1.3.0'
    ],
    scripts=[
        'scripts/rayvens-setup.sh', 'rayvens/rayvens', 'rayvens/rayvens_cli.py'
    ],
    version=version,
    python_requires='>=3.6',
    description='Rayvens augments Ray with events.',
    license='Apache 2.0',
    author='Rayvens authors',
    author_email='tardieu@us.ibm.com, gheorghe-teod.bercea@ibm.com',
    keywords=("ray events kubernetes"),
    url='https://github.com/project-codeflare/rayvens',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Java',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Operating System :: OS Independent',
        'Topic :: System :: Distributed Computing',
    ],
    project_urls={
        'Bug Reports': 'https://github.com/project-codeflare/rayvens/issues',
        'Source': 'https://github.com/project-codeflare/rayvens',
    },
)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re


# name if provided else kind
def name_source(config):
    if 'kind' not in config:
        raise TypeError('A Camel source needs a kind.')
    return config.get('name', config['kind'])


# name if provided else kind
def name_sink(config):
    if 'kind' not in config:
        raise TypeError('A Camel sink needs a kind.')
    return config.get('name', config['kind'])


# sanitize name for use as Kubernetes name and file name
def sanitize(name):
    name = str(name)
    name = name.lower()
    name = re.sub('[^a-z0-9]', '-', name)
    name = re.sub('-{2,}', '-', name)
    name = name[-253:]
    name = name.strip('-')
    return name


# combine stream name and source/sink name
def name_integration(stream_name, name_or_kind):
    return sanitize(stream_name + '-' + name_or_kind)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# The preloader file is fixed and can just be a string.
preloader_file_name = "Preloader.java"
preloader_file_contents = """
import org.apache.camel.BindToRegistry;
import org.apache.camel.Exchange;
import org.apache.camel.Processor;
import org.apache.camel.builder.RouteBuilder;

class Exit implements Processor {
  public void process(Exchange exchange) throws Exception {
    System.exit(0);
  }
}

public class Preloader extends RouteBuilder {
  @Override
  public void configure() throws Exception {
    from("timer:tick").to("bean:exit");
    from("platform-http:/null").to("http:null");
  }

  @BindToRegistry
  public Exit exit() {
    return new Exit();
  }
}
"""


def get_process_file_contents():
    return """
import java.nio.file.Path;
import java.nio.file.Paths;
import java.io.File;

import org.apache.camel.BindToRegistry;
import org.apache.camel.Exchange;
import org.apache.camel.Processor;
import org.apache.camel.builder.RouteBuilder;
import org.apache.camel.component.aws2.s3.AWS2S3Constants;

class FileProcessor implements Processor {
    public void process(Exchange exchange) throws Exception {
        // Input is a file. Set the header entry to the file name.
        File file = exchange.getIn().getBody(File.class);
        Path path = Paths.get(file.toString());
        exchange.getIn().setHeader(AWS2S3Constants.KEY, path.getFileName());
    }
}

public class ProcessFile extends RouteBuilder {
    @BindToRegistry
    public FileProcessor processFile() {
        return new FileProcessor();
    }

    @Override
    public void configure() throws Exception {
    }
}
    """


def get_process_path_contents():
    return """
import java.nio.file.Path;
import java.nio.file.Paths;

import org.apache.camel.BindToRegistry;
import org.apache.camel.Exchange;
import org.apache.camel.Processor;
import org.apache.camel.builder.RouteBuilder;
import org.apache.camel.component.aws2.s3.AWS2S3Constants;

class PathProcessor implements Processor {
    public void process(Exchange exchange) throws Exception {
        // Input is a string with the path to the file.
        Path path = Paths.get(exchange.getIn().getBody(String.class));
        exchange.getIn().setHeader(AWS2S3Constants.KEY, path.getFileName());
        exchange.getIn().setBody(path.toFile());
    }
}

public class ProcessPath extends RouteBuilder {
    @BindToRegistry
    public PathProcessor processPath() {
        return new PathProcessor();
    }

    @Override
    public void configure() throws Exception {
    }
}
    """


def get_job_launcher(launch_image):
    if launch_image is None:
        return ""

    job_command = "[env]"
    job_spec = f"""
        eventContents = eventContents.replace("\\\"", "");
        String newLine = System.getProperty("line.separator");
        String finalString = "apiVersion: batch/v1" + newLine
             + "kind: Job" + newLine
             + "metadata:" + newLine
             + "  name: "+ jobName + newLine
             + "spec:" + newLine
             + "  template:" + newLine
             + "    spec:" + newLine
             + "      containers:" + newLine
             + "      - name: " + jobName + "-container" + newLine
             + "        image: {launch_image}" + newLine
             + "        env:" + newLine
             + "        - name: TEST_EVENT" + newLine
             + "          value: \\\"" + eventContents + "\\\"" + newLine
             + "        command: {job_command}" + newLine
             + "      restartPolicy: Never" + newLine
             + "  backoffLimit: 4" + newLine;
        System.out.println(finalString);
        return finalString;
    """

    job_launcher = """
import java.lang.Runnable;
import java.lang.Thread;

import java.lang.ProcessBuilder;
import java.io.BufferedReader;

import java.io.InputStreamReader;
import java.io.FileWriter;
import java.util.function.Consumer;
import java.util.concurrent.Executors;
import java.lang.Runtime;
import java.lang.Process;

import java.io.IOException;
import com.google.gson.Gson;
import io.kubernetes.client.openapi.ApiClient;
import io.kubernetes.client.openapi.ApiException;
import io.kubernetes.client.openapi.apis.BatchV1Api;
import io.kubernetes.client.openapi.models.V1Job;
import io.kubernetes.client.util.Config;
import java.util.Map;
import org.yaml.snakeyaml.Yaml;
import org.yaml.snakeyaml.constructor.SafeConstructor;

class YamlRunnable implements Runnable {
    String body;
    int eventCount;

    public YamlRunnable(String body, int eventCount) {
        super();
        this.body = body;
        this.eventCount = eventCount;
    }

    public String getJobYaml(String jobName, String eventContents) {
        %s
    }

    public void createYamlFile(String filename) {
        try {
            File file = new File(filename);
            if (file.createNewFile()) {
                System.out.println("Yaml file created: " + file.getName());
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void writeJobYamlFile(String filename, String contents) {
        createYamlFile(filename);
        try {
            FileWriter myWriter = new FileWriter(filename);
            myWriter.write(contents);
            myWriter.close();
            System.out.println("Successfully wrote yaml file.");
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void deleteYamlFile(String filename) {
        File file = new File(filename);
        if (file.delete()) {
            System.out.println("Deleted the yaml file: " + file.getName());
        } else {
            System.out.println("Failed to delete the yaml file.");
        }
    }

    private static class StreamConsumer implements Runnable {
        private InputStream inputStream;
        private Consumer<String> consumer;

        public StreamConsumer(InputStream inputStream,
                              Consumer<String> consumer) {
            this.inputStream = inputStream;
            this.consumer = consumer;
        }

        @Override
        public void run() {
            new BufferedReader(new InputStreamReader(inputStream)).lines()
              .forEach(consumer);
        }
    }

    public void run() {
        System.out.println("Hello from a thread!");
        System.out.println(this.body);

        ProcessBuilder builder = new ProcessBuilder();
        String filename = "temp-job-"+String.valueOf(this.eventCount)+".yaml";
        String jobName = "random-job-name-"+String.valueOf(this.eventCount);
        writeJobYamlFile(filename, getJobYaml(jobName, this.body));
        builder.command("bash", "-c", "kubectl apply -f "+filename);

        try {
            Process process = builder.start();
            StreamConsumer outStream = new StreamConsumer(
                process.getInputStream(), System.out::println);
            Executors.newSingleThreadExecutor().submit(outStream);
            StreamConsumer errStream = new StreamConsumer(
                process.getErrorStream(), System.out::println);
            Executors.newSingleThreadExecutor().submit(errStream);
            try {
                int exitCode = process.waitFor();
                assert exitCode == 0;
            } catch (InterruptedException e) {
                System.err.println("Exception when calling process.waitFor()");
                e.printStackTrace();
            }
        } catch (IOException e) {
            System.err.println("Exception when calling builder.start()");
            e.printStackTrace();
        }

        deleteYamlFile(filename);
    }
}
    """ % (job_spec)
    return job_launcher


def get_queue_management_code(job_launcher, receive_processor, send_processor,
                              class_name, recv_method_name, send_method_name):
    return """
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.io.File;
import java.io.InputStream;
import java.io.FileInputStream;

import org.apache.camel.BindToRegistry;
import org.apache.camel.Exchange;
import org.apache.camel.Processor;
import org.apache.camel.builder.RouteBuilder;
import org.json.simple.JSONObject;

%s

class Recv implements Processor {
    BlockingQueue<Object> queue;
    int eventCount;

    public Recv(BlockingQueue<Object> queue) {
        this.queue = queue;
        this.eventCount = 0;
    }

    %s
}

class Send implements Processor {
    BlockingQueue<Object> queue;

    public Send(BlockingQueue<Object> queue) {
        this.queue = queue;
    }

    %s
}

public class %s extends RouteBuilder {
    BlockingQueue<Object> queue = new LinkedBlockingQueue<Object>();

    @BindToRegistry
    public Recv %s() {
        return new Recv(queue);
    }

    @BindToRegistry
    public Send %s() {
        return new Send(queue);
    }

    @Override
    public void configure() throws Exception {
    }
}
    """ % (job_launcher, receive_processor, send_processor, class_name,
           recv_method_name, send_method_name)


def job_launched_and_handler(launch_image):
    job_launcher = get_job_launcher(launch_image)
    handle_event_content = """
        queue.add(body);
    """
    if launch_image is not None:
        handle_event_content = """
        this.eventCount += 1;
        (new Thread(new YamlRunnable(body, this.eventCount))).start();
        """
    return job_launcher, handle_event_content


def get_java_file_queue_contents(launch_image):
    job_launcher, handle_event_content = job_launched_and_handler(launch_image)
    receive_processor = """
    public void process(Exchange exchange) throws Exception {
        File file = exchange.getIn().getBody(File.class);
        long fileSize = file.length();

        // TODO: find a better way to read the file content.
        byte[] allBytes = new byte[(int) fileSize];
        InputStream inputStream = new FileInputStream(file);
        inputStream.read(allBytes);
        inputStream.close();
        String body = new String(allBytes, StandardCharsets.UTF_8);
        %s
    }
    """ % (handle_event_content)

    send_processor = """
    public void process(Exchange exchange) throws Exception {
        Object body = queue.take();
        exchange.getIn().setBody(body);
    }
    """

    class_name = "FileQueue"
    recv_method_name = "addToFileQueue"
    send_method_name = "takeFromFileQueue"
    return get_queue_management_code(job_launcher, receive_processor,
                                     send_processor, class_name,
                                     recv_method_name, send_method_name)


def get_java_file_queue_json_contents(launch_image):
    job_launcher, handle_event_content = job_launched_and_handler(launch_image)
    receive_processor = """
    public void process(Exchange exchange) throws Exception {
        JSONObject returnJsonObject = new JSONObject();
        String body = exchange.getIn().getBody(String.class);
        returnJsonObject.put("body", body);

        Object key = exchange.getIn().getHeader("CamelAwsS3Key");
        returnJsonObject.put("filename", key.toString());
        String body = returnJsonObject.toString();
        %s
    }
    """ % (handle_event_content)

    send_processor = """
    public void process(Exchange exchange) throws Exception {
        Object body = queue.take();
        exchange.getIn().setBody(body);
    }
    """

    class_name = "FileQueueJson"
    recv_method_name = "addToFileJsonQueue"
    send_method_name = "takeFromFileJsonQueue"
    return get_queue_management_code(job_launcher, receive_processor,
                                     send_processor, class_name,
                                     recv_method_name, send_method_name)


def get_java_file_watch_queue_contents(launch_image):
    job_launcher, handle_event_content = job_launched_and_handler(launch_image)
    receive_processor = """
    public void process(Exchange exchange) throws Exception {
        JSONObject returnJsonObject = new JSONObject();

        // Record event type:
        Object eventType = exchange.getIn().getHeader("CamelFileEventType");
        returnJsonObject.put("event_type", eventType.toString());

        // Record event type:
        File file = exchange.getIn().getBody(File.class);
        returnJsonObject.put("filename", file.toString());
        String body = returnJsonObject.toString();
        %s
    }
    """ % (handle_event_content)

    send_processor = """
    public void process(Exchange exchange) throws Exception {
        Object body = queue.take();
        exchange.getIn().setBody(body);
    }
    """

    class_name = "FileWatchQueue"
    recv_method_name = "addToFileWatchQueue"
    send_method_name = "takeFromFileWatchQueue"
    return get_queue_management_code(job_launcher, receive_processor,
                                     send_processor, class_name,
                                     recv_method_name, send_method_name)


def get_java_meta_event_queue_contents(launch_image):
    job_launcher, handle_event_content = job_launched_and_handler(launch_image)
    receive_processor = """
    public void process(Exchange exchange) throws Exception {
        JSONObject returnJsonObject = new JSONObject();
        Object key = exchange.getIn().getHeader("CamelAwsS3Key");
        returnJsonObject.put("filename", key.toString());
        String body = returnJsonObject.toString();
        %s
    }
    """ % (handle_event_content)

    send_processor = """
    public void process(Exchange exchange) throws Exception {
        Object body = queue.take();
        exchange.getIn().setBody(body);
    }
    """

    class_name = "MetaEventQueue"
    recv_method_name = "addToMetaEventQueue"
    send_method_name = "takeFromMetaEventQueue"
    return get_queue_management_code(job_launcher, receive_processor,
                                     send_processor, class_name,
                                     recv_method_name, send_method_name)


def get_java_queue_contents(launch_image):
    job_launcher, handle_event_content = job_launched_and_handler(launch_image)
    receive_processor = """
    public void process(Exchange exchange) throws Exception {
        String body = exchange.getIn().getBody(String.class);
        %s
    }
    """ % (handle_event_content)

    send_processor = """
    public void process(Exchange exchange) throws Exception {
        Object body = queue.take();
        exchange.getIn().setBody(body);
    }
    """

    class_name = "Queue"
    recv_method_name = "addToQueue"
    send_method_name = "takeFromQueue"
    java_queue_file = get_queue_management_code(job_launcher,
                                                receive_processor,
                                                send_processor, class_name,
                                                recv_method_name,
                                                send_method_name)
    return java_queue_file



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
from rayvens.core import utils
from rayvens.core import kubernetes_utils

# Port used by the Quarkus Runtime to listen to HTTP requests.
quarkus_listener_port = "8080"

# Wait for pod to reach running state.


def pod_running_status(integration_name, namespace):
    # TODO: adapt this to support multiple namespaces.
    command = ["get", "pods", "-w"]

    # Namespace
    if namespace != "default":
        command.append("-n")
        command.append(namespace)

    return kubernetes_utils.pod_status(command, integration_name)


# Wait for integration to reach running state.


def integration_status(mode, pod_name, message=None):
    # TODO: adapt this to support multiple namespaces.
    command = ["logs", pod_name]

    # Namespace
    command.append("-n")
    command.append(mode.namespace)

    # Stream output from this command.
    command.append("--follow=true")

    return kubernetes_utils.invoke_kubectl_command(command,
                                                   message=message,
                                                   ongoing=True)


# Create service that Ray can talk to from outside the cluster.


def create_kamel_external_service(mode, service_name, integration_name):
    # Compose yaml file.
    yaml_file = f"""
kind: Service
apiVersion: v1
metadata:
  name: {service_name}
spec:
  ports:
  - nodePort: {utils.externalized_cluster_port}
    port: {quarkus_listener_port}
    protocol: TCP
    targetPort: {quarkus_listener_port}
  selector:
    camel.apache.org/integration: {integration_name}
  type: NodePort
    """

    # Write to output yaml file.
    output_file_name = os.path.abspath(service_name + ".yaml")
    output_file = open(output_file_name, "w")
    output_file.write(yaml_file)
    output_file.close()

    # Start service and check that it has started.
    service_started = False
    command = ["apply", "-f", output_file_name]

    # Namespace
    command.append("-n")
    command.append(mode.namespace)

    if kubernetes_utils.invoke_kubectl_command(command,
                                               service_name=service_name):
        command = ["get", "services", "-w"]

        # Namespace
        command.append("-n")
        command.append(mode.namespace)

        service_started = kubernetes_utils.invoke_kubectl_command(
            command, service_name=service_name, ongoing=True)

    if service_started:
        print("Service %s has been started successfully." % service_name)

    # Remove intermediate file.
    if output_file_name is not None:
        os.remove(output_file_name)


# Create service from yaml configuration:


def deploy_image_as_service(service_name, namespace, configuration_path):
    # TODO: adapt this to support multiple namespaces.
    command = ["apply"]

    # Namespace
    if namespace != "default":
        command.append("-n")
        command.append(namespace)

    # Service configuration
    command.append("-f")
    command.append(configuration_path)

    service_started = kubernetes_utils.invoke_kubectl_command(
        command, service_name=service_name, with_output=True)

    pod_running_status(service_name, namespace)

    if service_started:
        print("Service %s has been started successfully." % service_name)


# Delete service.


def delete_service(mode, service_name):
    command = ["delete", "service", service_name]

    # Namespace
    command.append("-n")
    command.append(mode.namespace)

    return kubernetes_utils.invoke_kubectl_command(command,
                                                   service_name=service_name)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from enum import Enum
from rayvens.core import utils

# Port for internal communication inside the cluster.
internal_cluster_port = "80"

# Cluster source port.
internal_cluster_port_for_source = "8000"


class RayvensMode(Enum):
    # Ray and Kamel running locally.
    LOCAL = 1

    # Ray running locally, Kamel local running in a container in the cluster.
    MIXED_LOCAL = 2

    # Ray running locally, Kamel operator running in the cluster.
    MIXED_OPERATOR = 3

    # Ray in cluster, Kamel local running in a container in the cluster.
    CLUSTER_LOCAL = 4

    # Ray in cluster, Kamel operator running in the cluster.
    CLUSTER_OPERATOR = 5

    # Ray operator in cluster, Kamel local running in a container in cluster.
    OPERATOR_LOCAL = 6

    # Ray operator in cluster, Kamel operator running in the cluster.
    OPERATOR_OPERATOR = 7


class RunMode:
    def __init__(self, run_mode=RayvensMode.LOCAL):
        self.run_mode = run_mode
        self.namespace = "ray"
        self.transport = None
        self.check_port = True
        self.release = False

    def server_address(self, integration):
        return self._get_server_address(integration.integration_name,
                                        port=integration.port)

    def is_local(self):
        return self.run_mode == RayvensMode.LOCAL

    def is_mixed(self):
        return self.run_mode == RayvensMode.MIXED_OPERATOR

    def is_cluster(self):
        return self.run_mode == RayvensMode.CLUSTER_OPERATOR

    def _get_server_address(self,
                            integration_name,
                            serve_source=False,
                            port=None):
        if self.run_mode == RayvensMode.LOCAL:
            # Default setup: "http://0.0.0.0:8080"
            if port is None:
                raise RuntimeError('port is not specified')
            return f'http://localhost:{port}'
        if self.run_mode == RayvensMode.MIXED_OPERATOR:
            return "http://localhost:%s" % utils.externalized_cluster_port
        if self.run_mode == RayvensMode.CLUSTER_OPERATOR:
            if integration_name == "":
                raise RuntimeError("integration name is not set")
            if serve_source:
                return "http://%s.%s.svc.cluster.local:%s" % (
                    integration_name, self.namespace,
                    internal_cluster_port_for_source)
            return "http://%s.%s.svc.cluster.local:%s" % (
                integration_name, self.namespace, internal_cluster_port)
        raise RuntimeError("unreachable")


# Default execution mode.
mode = RunMode()



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from enum import Enum
from rayvens.core import invocation


class KamelCommand(Enum):
    INSTALL = 1
    BUILD = 2
    RUN = 3
    LOCAL_BUILD = 4
    LOCAL_RUN = 5
    DELETE = 6
    LOG = 7
    UNINSTALL = 100


def kamel_command_type(command_options):
    command = ' '.join(command_options)
    if command.startswith("install"):
        return KamelCommand.INSTALL
    if command.startswith("build"):
        return KamelCommand.BUILD
    if command.startswith("run"):
        return KamelCommand.RUN
    if command.startswith("local build"):
        return KamelCommand.LOCAL_BUILD
    if command.startswith("local run"):
        return KamelCommand.LOCAL_RUN
    if command.startswith("uninstall"):
        return KamelCommand.UNINSTALL
    if command.startswith("delete"):
        return KamelCommand.DELETE
    if command.startswith("log"):
        return KamelCommand.LOG
    raise RuntimeError('unsupported kamel subcommand: %s' % command)


def kamel_command_str(command_type):
    if command_type == KamelCommand.INSTALL:
        return "install"
    if command_type == KamelCommand.BUILD:
        return "build"
    if command_type == KamelCommand.RUN:
        return "run"
    if command_type == KamelCommand.LOCAL_BUILD:
        return "local build"
    if command_type == KamelCommand.LOCAL_RUN:
        return "local run"
    if command_type == KamelCommand.UNINSTALL:
        return "uninstall"
    if command_type == KamelCommand.DELETE:
        return "delete"
    if command_type == KamelCommand.LOG:
        return "log"
    raise RuntimeError('unsupported kamel subcommand')


def kamel_command_end_condition(subcommand_type, base_name):
    if subcommand_type == KamelCommand.INSTALL:
        return "Camel K installed in namespace"
    if subcommand_type == KamelCommand.BUILD:
        return ""
    if subcommand_type == KamelCommand.RUN:
        return f"Integration \"{base_name}\""
    if subcommand_type == KamelCommand.LOCAL_BUILD:
        return ""
    if subcommand_type == KamelCommand.LOCAL_RUN:
        return "Installed features:"
    if subcommand_type == KamelCommand.UNINSTALL:
        return "Camel K Service Accounts removed from namespace"
    if subcommand_type == KamelCommand.DELETE:
        return f"Integration {base_name} deleted"
    if subcommand_type == KamelCommand.LOG:
        return ""
    raise RuntimeError('unsupported kamel subcommand: %s' %
                       kamel_command_str(subcommand_type))


# Helper for ongoing local commands like kamel local run.


def invoke_kamel_command(command,
                         mode,
                         integration_name,
                         integration_content=[],
                         message=None):
    # Invoke command using the Kamel invocation actor.
    kamel_invocation = invocation.KamelInvocation(
        command,
        mode,
        integration_name=integration_name,
        integration_content=integration_content)

    # Wait for kamel command to finish launching the integration.
    if kamel_invocation.invoke(message):
        return kamel_invocation

    return None



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import ray
import time
from rayvens.core import kamel


def verify_do(handle, _global_camel, action, *args, **kwargs):
    if action == 'verify_log':
        return _verify_log(handle, _global_camel, *args, **kwargs)
    raise RuntimeError('invalid meta action')


def wait_for_event(handle):
    event_count = 0
    countdown = 20
    while event_count == 0:
        event_count = ray.get(handle.event_count.remote())
        time.sleep(1)
        countdown -= 1
        if countdown == 0:
            break
    if event_count == 0:
        return False
    return True


def _verify_log(handle,
                _global_camel,
                sink_source_name,
                message,
                wait_for_events=False):
    log = "FAIL"

    # Wait for at least one event to happen.
    if wait_for_events:
        if not wait_for_event(handle):
            print("[LOG CHECK]:", "NO EVENTS RECEIVED")
            return False

    if _global_camel.mode.is_local():
        # In the local case the integration run is ongoing and we can
        # access the logs directly.
        outcome = ray.get(
            handle._integration_invoke.remote(sink_source_name, message))
    else:
        # When running using the operator then the integration run command
        # is non-blocking and returns immediately. The logs can be queried
        # using the kamel log command.
        integration_name = ray.get(
            handle._get_integration_name.remote(sink_source_name))
        invocation = kamel.log(_global_camel.mode, integration_name, message)
        outcome = invocation is not None
        invocation.kill()

    if outcome:
        log = "SUCCESS"
    print("[LOG CHECK]:", log)
    return outcome
